import typing

from autonetkit.network_model.exceptions import NodeNotFound
from autonetkit.network_model.types import LinkId

if typing.TYPE_CHECKING:
    from autonetkit.network_model.port import Port
    from autonetkit.network_model.node import Node
    from autonetkit.network_model.topology import Topology


class Link:
    """

    """

    def __init__(self, topology, id, p1: 'Port', p2: 'Port'):
        self.topology: Topology = topology
        self.id: LinkId = id
        self.p1: Port = p1
        self.p2: Port = p2
        self._data = {}

    @property
    def local_data(self) -> typing.Dict:
        """

        @return:
        """
        return self._data

    def __repr__(self):
        return f"L {self.p1}.{self.n1} {self.p1}.{self.n2}"

    def set(self, key, val) -> None:
        """

        @param key:  The key to set
        @param val:  The value to set
        """
        self._data[key] = val

    def get(self, key, default=None):
        """

        @param key: the key to get
        @param default:
        @return:
        """
        try:
            return self._data[key]
        except KeyError:
            return default

    def export(self) -> typing.Dict:
        """

        @return:
        """
        data = self._data.copy()
        data["p1"] = self.p1.id
        data["p2"] = self.p2.id
        data["n1"] = self.p1.node.id
        data["n2"] = self.p2.node.id
        return data

    @property
    def n1(self) -> 'Node':
        """

        @return:
        """
        return self.p1.node

    @property
    def n2(self) -> 'Node':
        """

        @return:
        """
        return self.p2.node

    def other_node(self, node: 'Node') -> 'Node':
        """

        @param node:
        @return: the other node to the parameter
        """
        if self.n1 == node:
            return self.n2
        elif self.n2 == node:
            return self.n1
        else:
            raise NodeNotFound(node)



import logging

import aiohttp
import aiohttp_jinja2
import jinja2
import pkg_resources
from aiohttp import web

logger = logging.getLogger(__name__)


def main():
    """

    @return:
    """

    async def websocket_handler(request):
        """

        @param request:
        @return:
        """
        ws = web.WebSocketResponse()
        request.app.websockets.add(ws)
        await ws.prepare(request)
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    if msg.data == 'close':
                        await ws.close()
                        request.app.websockets.remove(ws)
                    else:
                        await ws.send_str("Connected")
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    logger.warning('ws connection closed with exception %s' %
                                   ws.exception())

        finally:
            request.app.websockets.remove(ws)
        return ws

    @aiohttp_jinja2.template('index.jinja2')
    async def index(request):
        """

        @param request:
        @return:
        """
        if request.app.data is None:
            return web.Response(text="No topology set")

        topology_name = request.query.get("topology")
        topology_names = list(request.app.data.keys())

        try:
            topology = request.app.data[topology_name]
        except KeyError:
            topology = request.app.data["physical"]

        return {'topology': topology, 'topology_names': topology_names, "topology_name": topology_name}

    async def hello(request):
        """

        @param request:
        @return:
        """
        return web.json_response(request.app.data)

    async def post_handler(request):
        """

        @param request:
        @return:
        """
        data = await request.json()
        request.app.data = data
        for ws in request.app.websockets:
            await ws.send_str("Updated")

        return web.Response(text="Data stored")

    app = web.Application()
    app.data = None
    app.websockets = set()
    app.add_routes([web.get('/', index)])
    app.add_routes([web.get('/data', hello)])
    app.add_routes([web.post('/data', post_handler)])
    app.add_routes([web.get('/ws', websocket_handler)])

    aiohttp_jinja2.setup(app,
                         loader=jinja2.FileSystemLoader('templates'))

    resource_package = __name__
    resource_path = 'static'
    path_to_static_folder = pkg_resources.resource_filename(resource_package, resource_path)

    app.add_routes([web.static('/static', path_to_static_folder)])

    web.run_app(app)


if __name__ == '__main__':
    main()



import sys

from autonetkit.workflow.workflow import BaseWorkflow


def main(filename):
    """

    @param filename:
    """
    workflow = BaseWorkflow()
    network_model = workflow.load(filename)
    workflow.run(network_model, target_platform="kathara")


if __name__ == '__main__':
    filename = sys.argv[1]
    main(filename)



import ipaddress
from collections import defaultdict

from autonetkit.design.utils import filters
from autonetkit.design.utils.general import group_by
from autonetkit.network_model.types import LAYER3_DEVICES


def assign_loopbacks(topology):
    """

    @param topology:
    """
    layer3_nodes = [n for n in topology.nodes()
                    if n.type in LAYER3_DEVICES]
    loopback_block = ipaddress.IPv4Network("172.16.0.0/16")
    loopback_subnets = loopback_block.subnets(new_prefix=24)
    grouped_l3 = group_by(layer3_nodes, "asn")

    allocated_loopbacks = defaultdict(list)

    for asn, nodes in grouped_l3.items():
        # can repeat the loopbacks in each asn
        subnet = next(loopback_subnets)
        allocated_loopbacks[asn].append(subnet)
        host_ips = subnet.hosts()
        for node in nodes:
            host_ip = next(host_ips)
            lo0 = node.loopback_zero()
            lo0.set("ip", host_ip)
            # also map onto node for debugging/vis

    topology.set("loopbacks_by_asn", allocated_loopbacks)


def assign_bc_subnets(topology):
    """

    @param topology:
    """
    # the network to use to address end hosts and for inter-domain connections
    allocated_blocks = defaultdict(list)
    global_advertise_network = ipaddress.IPv4Network("10.0.0.0/8")
    global_subnets = global_advertise_network.subnets(new_prefix=16)
    bc_nodes = filters.broadcast_domains(topology)
    grouped_bc = group_by(bc_nodes, "asn")
    for asn, nodes in grouped_bc.items():
        asn_block = next(global_subnets)
        allocated_blocks[asn].append(asn_block)
        # quick method: allocate a /24 to each broadcast domain
        # Note: this could be significantly optimised in the future
        # Note: could allocate different block to internal infrastructure too
        external_blocks = asn_block.subnets(new_prefix=24)
        for bc in nodes:
            bc_block = next(external_blocks)
            bc.set("network", bc_block)
    topology.set("infrastructure_by_asn", allocated_blocks)



from autonetkit.packager.base import AbstractPlatformPackager
from autonetkit.packager.kathara import KatharaPlatformPackager


def get_platform_builder(target: str) -> AbstractPlatformPackager:
    """

    @param target:
    @return:
    """
    map = {
        "kathara": KatharaPlatformPackager()
    }

    return map[target]



from autonetkit.design.layer2 import build_l2, build_l2_conn
from autonetkit.design.routing import build_ospf, build_ebgp, build_ibgp
from autonetkit.design.validation.validation import check_layer2_conn
from autonetkit.network_model.network_model import NetworkModel


class Builder:
    """

    """

    def __init__(self, network_model: NetworkModel):
        self.network_model = network_model

    def build(self) -> None:
        """

        """
        self.build_physical()
        self.build_layer2()
        self.build_ipv4()
        self.build_routing()

    def build_physical(self):
        """

        """
        t_in = self.network_model.get_topology("input")
        t_phy = self.network_model.get_topology("physical")
        t_phy.add_nodes_from(t_in.nodes())
        t_phy.add_links_from(t_in.links())

    def build_layer2(self):
        """

        """
        build_l2(self.network_model)
        build_l2_conn(self.network_model)
        t_l2_conn = self.network_model.get_topology("layer2_conn")
        check_layer2_conn(t_l2_conn)

    def build_ipv4(self):
        """

        """
        import autonetkit.design.ip_addressing.ipv4 as ipv4
        import autonetkit.design.ip_addressing.general as general
        g_ipv4 = self.network_model.create_topology("ipv4")
        general.build_ip_base(self.network_model, g_ipv4)
        general.allocate_asns(g_ipv4)
        ipv4.assign_bc_subnets(g_ipv4)
        general.map_blocks_to_ports(g_ipv4)
        ipv4.assign_loopbacks(g_ipv4)

    def build_routing(self):
        """

        """
        build_ospf(self.network_model)
        build_ibgp(self.network_model)
        build_ebgp(self.network_model)



import unittest

from autonetkit.entry import main


class MyTestCase(unittest.TestCase):
    @staticmethod
    def test_something():
        filename = "../example/small_internet.graphml"
        main(filename)


if __name__ == '__main__':
    unittest.main()



import statistics

import typing
from collections import defaultdict

if typing.TYPE_CHECKING:
    from autonetkit.network_model.node import Node


def move_to_average_peer_locations(node: 'Node') -> None:
    """

    @param node:
    """
    all_x = [n.get("x") for n in node.peer_nodes() if n.get("x") is not None]
    all_y = [n.get("y") for n in node.peer_nodes() if n.get("y") is not None]
    if len(all_x) and len(all_y):
        mean_x = statistics.mean(all_x)
        mean_y = statistics.mean(all_y)
    else:
        mean_x = mean_y = None
    node.set("x", mean_x)
    node.set("y", mean_y)


def group_by(nodes: typing.List['Node'], key: str) -> typing.Dict[str, typing.List['Node']]:
    """

    @param nodes:
    @param key:
    @return:
    """
    result = defaultdict(list)
    for node in nodes:
        result[node.get(key)].append(node)

    return dict(result)



import logging
import typing
from collections import defaultdict
from typing import List, Dict

from autonetkit.network_model.exceptions import NodeNotFound, LinkNotFound, PortNotFound, PortIdInUse, LinkIdInUse, \
    NodeIdInUse, PathIdInUse, NumericNodeId
from autonetkit.network_model.link import Link
from autonetkit.network_model.node import Node
from autonetkit.network_model.path import NodePath, PortPath
from autonetkit.network_model.port import Port
from autonetkit.network_model.types import DeviceType, NodeId, TopologyId, LinkId, PortId, PortType, PathId

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from autonetkit.network_model.network_model import NetworkModel


class Topology:
    """

    """

    def __init__(self, network_model, id: TopologyId):
        self.id: TopologyId = id
        self.network_model: NetworkModel = network_model
        self._nodes: Dict[NodeId: Node] = {}
        self._links: Dict[LinkId: Link] = {}
        self._ports: Dict[PortId, Port] = {}
        self._node_paths: Dict[PathId, NodePath] = {}
        self._port_paths: Dict[PathId, PortPath] = {}

        self._cache_ports_for_node: Dict[NodeId, List[Port]] = defaultdict(list)
        self._cache_links_for_node: Dict[NodeId, List[Link]] = defaultdict(list)

        self._data = {}

    def create_node(self, type: DeviceType, label: typing.Optional[str] = None,
                    id: typing.Optional[NodeId] = None) -> 'Node':
        """

        @param type:
        @param label:
        @param id:
        @return:
        """
        if id is not None:
            if id in self.network_model.used_node_ids:
                raise NodeIdInUse(id)
            else:
                self.network_model.used_node_ids.add(id)
                node_id = id
        else:
            node_id = self.network_model.generate_node_id()

        node = self._create_node(node_id)
        node.set("type", type)
        if label:
            node.set("label", label)

        return node

    def _create_node(self, node_id):
        node = Node(self, node_id)
        self._nodes[node_id] = node
        return node

    def create_port(self, node: 'Node', type: PortType, label: typing.Optional[str] = None,
                    id: typing.Optional[PortId] = None) -> 'Port':
        """

        @param node:
        @param type:
        @param label:
        @param id:
        @return:
        """
        if id is not None:
            if id in self.network_model.used_port_ids:
                raise PortIdInUse(id)
            else:
                self.network_model.used_port_ids.add(id)
                port_id = id
        else:
            port_id = self.network_model.generate_port_id()
        port = self._create_port(node, port_id)
        port.set("type", type)
        if label:
            port.set("label", label)

        return port

    def _create_port(self, node, port_id):
        port = Port(node, port_id)
        self._cache_ports_for_node[port.node.id].append(port)
        self._ports[port_id] = port
        return port

    def create_link(self, p1: 'Port', p2: 'Port',
                    id: typing.Optional[LinkId] = None) -> 'Link':
        """

        @param p1:
        @param p2:
        @param id:
        @return:
        """
        if id is not None:
            if id in self.network_model.used_link_ids:
                raise LinkIdInUse(id)
            else:
                self.network_model.used_link_ids.add(id)
                link_id = id
        else:
            link_id = self.network_model.generate_link_id()
        link = self._create_link(link_id, p1, p2)

        return link

    def _create_link(self, link_id, p1, p2):
        # TODO: warn if port types are not the same
        link = Link(self, link_id, p1, p2)
        self._links[link_id] = link
        self._cache_links_for_node[link.n1.id].append(link)
        self._cache_links_for_node[link.n2.id].append(link)
        return link

    def create_node_path(self, nodes: List[Node], id=None) -> NodePath:
        """

        @param nodes:
        @param id:
        @return:
        """
        if id is not None:
            if id in self.network_model.used_path_ids:
                raise PathIdInUse(id)
            else:
                self.network_model.used_path_ids.add(id)
                path_id = id
        else:
            path_id = self.network_model.generate_path_id()
        path = NodePath(self, path_id, nodes)
        self._node_paths[path_id] = path
        return path

    def create_port_path(self, ports: List[Port], id=None) -> PortPath:
        """

        @param ports:
        @param id:
        @return:
        """
        if id is not None:
            if id in self.network_model.used_path_ids:
                raise PathIdInUse(id)
            else:
                self.network_model.used_path_ids.add(id)
                path_id = id
        else:
            path_id = self.network_model.generate_path_id()
        path = PortPath(self, path_id, ports)

        self._port_paths[path_id] = path
        return path

    def add_nodes_from(self, nodes: List['Node'], with_ports: bool = True,
                       ports: typing.Optional[typing.Iterable['Port']] = None):
        """

        @param nodes:
        @param with_ports:
        @param ports:
        """
        if ports:
            ports = set(ports)

        for original_node in nodes:
            node_id = original_node.id
            node = self._create_node(node_id)

            ports_to_copy = []
            if ports:
                ports_to_copy = set(node.ports()).intersection(ports)

            elif with_ports:
                ports_to_copy = original_node.ports()

            for port in ports_to_copy:
                self._create_port(node, port.id)

    def add_links_from(self, links: List["Link"], raise_if_port_not_found=False):
        """

        @param links:
        @param raise_if_port_not_found:
        """
        for link in links:
            try:
                p1 = self.get_port_by_id(link.p1.id)
                p2 = self.get_port_by_id(link.p2.id)
            except PortNotFound:
                if raise_if_port_not_found:
                    raise
                else:
                    continue

            self._create_link(link.id, p1, p2)

    def remove_nodes_from(self, nodes: List['Node']) -> None:
        """

        @param nodes:
        """
        for node in nodes:
            self.remove_node(node)

    def remove_links_from(self, links: List['Link']) -> None:
        """

        @param links:
        """
        for link in links:
            self.remove_link(link)

    def remove_node(self, node: 'Node') -> None:
        """

        @param node:
        """
        links = list(node.links())
        for link in links:
            self.remove_link(link)

        ports = list(node.ports())
        for port in ports:
            self.remove_port(port)

        try:
            del self._nodes[node.id]
        except KeyError:
            pass  # already removed

        try:
            del self._cache_ports_for_node[node.id]
            del self._cache_links_for_node[node.id]
        except KeyError:
            pass  # already removed

    def remove_link(self, link: 'Link') -> None:
        """

        @param link:
        """
        # TODO: also allow to remove ports too?
        try:
            del self._links[link.id]
        except KeyError:
            pass  # already removed

        try:
            self._cache_links_for_node[link.n1.id].remove(link)
        except ValueError:
            pass  # already removed

        try:
            self._cache_links_for_node[link.n2.id].remove(link)
        except ValueError:
            pass  # already removed

    def remove_port(self, port: 'Port') -> None:
        """

        @param port:
        """
        links = list(port.links())
        for link in links:
            # TODO: check this is correctly capturing the nodes/links
            self.remove_link(link)

        try:
            del self._ports[port.id]
        except KeyError:
            pass  # already removed

        try:
            self._cache_ports_for_node[port.node.id].remove(port)
        except ValueError:
            pass  # already removed

    def remove_node_path(self, path: 'NodePath') -> None:
        """

        @param path:
        """
        try:
            del self._node_paths[path.id]
        except KeyError:
            pass  # already removed

    def remove_port_path(self, path: 'PortPath') -> None:
        """

        @param path:
        """
        try:
            del self._port_paths[path.id]
        except KeyError:
            pass  # already removed

    def set(self, key, val) -> None:
        """

        @param key:
        @param val:
        """
        self._data[key] = val

    def get(self, key, default=None):
        """

        @param key:
        @param default:
        @return:
        """
        try:
            return self._data[key]
        except KeyError:
            return default

    def nodes(self) -> typing.List['Node']:
        """

        @return:
        """
        return list(self._nodes.values())

    def ports(self, node: typing.Optional['Node'] = None) -> typing.List['Port']:
        """

        @param node:
        @return:
        """
        if node:
            return self._cache_ports_for_node[node.id]
        else:
            return list(self._ports.values())

    def links(self, node: typing.Optional['Node'] = None) -> typing.List['Link']:
        """

        @param node:
        @return:
        """
        if node:
            return self._cache_links_for_node[node.id]
        else:
            return list(self._links.values())

    def node_paths(self) -> typing.List['NodePath']:
        """

        @return:
        """
        return list(self._node_paths.values())

    def port_paths(self) -> typing.List['PortPath']:
        """

        @return:
        """
        return list(self._port_paths.values())

    def export(self) -> Dict:
        """

        @return:
        """
        result = {
            "nodes": {},
            "links": {},
            "ports": {},
            "node_paths": {},
            "port_paths": {}
        }

        for node_id, node in self._nodes.items():
            result["nodes"][node_id] = node.export()

        for link_id, link in self._links.items():
            result["links"][link_id] = link.export()

        for port_id, port in self._ports.items():
            result["ports"][port_id] = port.export()

        for path_id, path in self._node_paths.items():
            result["node_paths"][path_id] = path.export()

        for path_id, path in self._port_paths.items():
            result["port_paths"][path_id] = path.export()

        return result

    def get_node_by_id(self, nid: NodeId) -> 'Node':
        """

        @param nid:
        @return:
        """
        try:
            return self._nodes[nid]
        except KeyError:
            if isinstance(nid, int):
                # may be from networkx
                message = f"Attempting to search for node by numeric id, cast to a string {nid}"
                raise NumericNodeId(message)

            raise NodeNotFound(nid)

    def get_link_by_id(self, lid: LinkId) -> 'Link':
        """

        @param lid:
        @return:
        """
        try:
            return self._links[lid]
        except KeyError:
            raise LinkNotFound(lid)

    def get_port_by_id(self, pid: PortId) -> 'Port':
        """

        @param pid:
        @return:
        """
        try:
            return self._ports[pid]
        except KeyError:
            raise PortNotFound(pid)



import logging
from collections import Counter

from autonetkit.design.utils import filters
from autonetkit.network_model.exceptions import PortNotFound
from autonetkit.network_model.topology import Topology
from autonetkit.network_model.types import LAYER3_DEVICES

logger = logging.getLogger(__name__)


def validate(topology: Topology) -> bool:
    """
    TODO: can also check no two hosts directly connected, and only routers cross ASNs


    """
    valid = True
    if not labels_unique(topology):
        valid = False

    if not physical_ports_have_slot(topology):
        valid = False

    if not l3_nodes_have_loopback(topology):
        valid = False

    return valid


def physical_ports_have_slot(topology) -> bool:
    """

    @param topology:
    @return:
    """
    valid = True
    for node in topology.nodes():
        for port in filters.physical_ports(node):
            if port.slot is None:
                logger.warning("No slot allocated for %s on %s", port.label, node.label)
                valid = False

    return valid


def l3_nodes_have_loopback(topology) -> bool:
    """

    @param topology:
    @return:
    """
    valid = True
    layer3_nodes = [n for n in topology.nodes()
                    if n.type in LAYER3_DEVICES]
    for node in layer3_nodes:
        try:
            _ = node.loopback_zero()
        except PortNotFound:
            logger.warning("Loopback zero not set for %s", node.label)
            valid = False

    return valid


def labels_unique(topology) -> bool:
    """

    @param topology:
    @return:
    """
    valid = True
    label_counter = Counter()
    for node in topology.nodes():
        label_counter[node.label] += 1
    for label, count in label_counter.items():
        if count > 1:
            logger.warning("Label %s is present %s times", label, count)
            valid = False
    return valid



import itertools

from autonetkit.design.utils import filters
from autonetkit.design.utils.general import move_to_average_peer_locations
from autonetkit.design.utils.graph_utils import connected_components, merge_nodes, split_link, explode_node
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.types import DeviceType, LAYER3_DEVICES


def build_l2(network_model: NetworkModel):
    """

    @param network_model:
    """
    # TODO: check if should be called l2 or broadcast domains
    # TODO: later extend to support managed switches also
    t_phy = network_model.get_topology("physical")
    t_l2 = network_model.create_topology("layer2")

    # TODO: later can filter out hubs, optical gear, etc if included from l1 topology
    t_l2.add_nodes_from(t_phy.nodes())
    t_l2.add_links_from(t_phy.links())

    # merge switches
    switches = filters.switches(t_l2)
    components = connected_components(t_l2, switches)

    labels = (f"bc_{x}" for x in itertools.count())

    for component_nodes in components:
        label = next(labels)
        merged = merge_nodes(t_l2, component_nodes, label=label)
        move_to_average_peer_locations(merged)

        merged.set("type", DeviceType.BROADCAST_DOMAIN)

    # split router ptp links
    ptp_links = [l for l in t_l2.links()
                 if l.n1.type in LAYER3_DEVICES
                 and l.n2.type in LAYER3_DEVICES]

    for link in ptp_links:
        label = next(labels)
        split = split_link(t_l2, link, label)
        move_to_average_peer_locations(split)
        split.set("type", DeviceType.BROADCAST_DOMAIN)


def build_l2_conn(network_model):
    """

    @param network_model:
    """
    t_l2 = network_model.get_topology("layer2")
    t_l2_conn = network_model.create_topology("layer2_conn")

    t_l2_conn.add_nodes_from(t_l2.nodes())
    t_l2_conn.add_links_from(t_l2.links())
    bc_nodes = filters.broadcast_domains(t_l2_conn)

    for node in bc_nodes:
        explode_node(t_l2_conn, node)



from typing import Dict, List

from autonetkit.compile.node.base import BaseCompiler
from autonetkit.design.utils import filters
from autonetkit.network_model.exceptions import NodeNotFound
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.node import Node


class QuaggaCompiler(BaseCompiler):
    """

    """

    def __init__(self):
        super().__init__()

        # TODO: assign this at topology load time in pre-process step
        self.lo0_id = "lo:1"

    def compile(self, network_model: NetworkModel, node: Node) -> Dict:
        """

        @param network_model:
        @param node:
        @return:
        """
        # ensure node is from phy
        t_phy = network_model.get_topology("physical")
        node = t_phy.get_node_by_id(node.id)

        result = {}

        result["zebra"] = self._zebra(node)
        result["interfaces"] = self._interfaces(network_model, node)
        result["ospf"] = self._ospf(network_model, node)
        result["bgp"] = self._bgp(network_model, node)

        return result

    def _zebra(self, node):
        return {
            "hostname": node.label,
            "password": "zebra",
            "enable_password": "zebra",
        }

    def _interfaces(self, network_model: NetworkModel, node: Node) -> List:
        result = []

        t_ipv4 = network_model.get_topology("ipv4")

        physical_ports = filters.physical_ports(node)
        logical_ports = [p for p in node.ports()
                         if p not in physical_ports]

        physical_ports.sort(key=lambda x: x.label)
        logical_ports.sort(key=lambda x: x.label)
        for port in physical_ports:
            try:
                description = port.peer_ports()[0].node.label
            except IndexError:
                description = ""

            data = {
                "label": port.label,
                "description": description,
                "connected": port.connected
            }

            if port.connected:
                ipv4_port = t_ipv4.get_port_by_id(port.id)
                data["broadcast"] = ipv4_port.get("network").broadcast_address
                data["network"] = ipv4_port.get("network").network_address
                data["ip"] = ipv4_port.get("ip")

            result.append(data)

        return result

    def _ospf(self, network_model: NetworkModel, node: Node) -> Dict:
        t_ospf = network_model.get_topology("ospf")
        try:
            ospf_node = t_ospf.get_node_by_id(node.id)
        except NodeNotFound:
            # ospf not configured for this node
            return {}

        result = {}
        asn = node.get("asn")
        t_ipv4 = network_model.get_topology("ipv4")

        infrastructure_blocks = t_ipv4.get("infrastructure_by_asn").get(asn) or []
        networks = []
        for network in infrastructure_blocks:
            networks.append({
                "network": network.network_address,
                "area": ospf_node.get("area")
            })

        result["networks"] = networks

        return result

    def _bgp(self, network_model: NetworkModel, node: Node) -> Dict:
        result = {}
        asn = node.get("asn")
        result["asn"] = asn

        t_ipv4 = network_model.get_topology("ipv4")
        ipv4_node = t_ipv4.get_node_by_id(node.id)
        lo0_ipv4_address = ipv4_node.loopback_zero().get("ip")

        infrastructure_blocks = t_ipv4.get("infrastructure_by_asn").get(asn) or []

        result["networks"] = infrastructure_blocks

        t_ibgp = network_model.get_topology("ibgp")
        ibgp_node = t_ibgp.get_node_by_id(node.id)

        ibgp_neighbors = []
        for peer_port in ibgp_node.peer_ports():
            peer_node = peer_port.node
            peer_ip_port = t_ipv4.get_port_by_id(peer_port.id)
            peer_ipv4 = peer_ip_port.get("ip")
            neighbor = f"{peer_port.label}.{peer_node.label}"
            description = "link to " + neighbor
            ibgp_neighbors.append({
                "update_source": lo0_ipv4_address,
                "desc": neighbor,
                "asn": peer_node.get("asn"),
                "neigh_ip": peer_ipv4,
                "neighbor": neighbor,
                "description": description
            })

        result["ibgp_neighbors"] = ibgp_neighbors

        t_ebgp = network_model.get_topology("ebgp")
        ebgp_node = t_ebgp.get_node_by_id(node.id)

        ebgp_neighbors = []
        for peer_port in ebgp_node.peer_ports():
            peer_node = peer_port.node
            peer_ip_port = t_ipv4.get_port_by_id(peer_port.id)
            peer_ipv4 = peer_ip_port.get("ip")
            neighbor = f"{peer_port.label}.{peer_node.label}"
            description = "link to " + neighbor
            ebgp_neighbors.append({
                "update_source": lo0_ipv4_address,
                "desc": neighbor,
                "asn": peer_node.get("asn"),
                "neigh_ip": peer_ipv4,
                "neighbor": neighbor,
                "description": description
            })

        result["ebgp_neighbors"] = ebgp_neighbors

        return result



import abc
from typing import Dict

from jinja2 import Environment

from autonetkit.network_model.node import Node


class NodeRenderer:
    """

    """

    def __init__(self, env: Environment):
        self.env: Environment = env

    @abc.abstractmethod
    def render(self, node: Node) -> Dict:
        """

        @param node:
        """
        pass



import logging
import os

from autonetkit.compile.node_compiler import TargetCompilerNotFound, NodeCompiler
from autonetkit.compile.platform.kathara import KatharaCompiler
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.packager.base import AbstractPlatformPackager
from autonetkit.packager.utils import write_rendered_to_file
from autonetkit.render.exceptions import RenderException
from autonetkit.render.platform.kathara import KatharaRenderer
from autonetkit.render.renderer import Renderer

logger = logging.getLogger(__name__)


class KatharaPlatformPackager(AbstractPlatformPackager):
    """

    """

    def __init__(self):
        # TODO: allow passing in lab metadata (name, description, etc) info also
        super().__init__()

    def build(self, network_model: NetworkModel) -> None:
        """

        @param network_model:
        """
        compiler = NodeCompiler()
        renderer = Renderer()

        base_dir = os.path.join(self.output_dir, "lab")
        try:
            os.mkdir(base_dir)
        except FileExistsError:
            pass
        except FileNotFoundError as err:
            raise RenderException(err)

        self._compile_nodes(base_dir, compiler, network_model, renderer)
        self._compile_platform(base_dir, network_model, renderer)

    def _compile_nodes(self, base_dir, compiler, network_model, renderer):
        platform_compiler = KatharaCompiler()
        platform_renderer = KatharaRenderer(renderer.env)
        for node in network_model.get_topology("physical").nodes():
            try:
                node_data = compiler.compile_node(network_model, node)
            except TargetCompilerNotFound:
                logger.warning("No compiler defined for %s", node.label)
                continue

            target = node.get("target")
            rendered = renderer.render_node(node_data, target)

            # TODO: add hostname as platform specific extra renderer
            node_data["kathara"] = {
                "hostname": platform_compiler.compile_node(node)
            }

            # also append kathara specific files to allow ssh into node
            # Note: these don't need compilation
            rendered += platform_renderer.render_node(node_data)

            node_dir = os.path.join(base_dir, node.label)
            try:
                os.mkdir(node_dir)
            except FileExistsError:
                pass

            for entry in rendered:
                write_rendered_to_file(entry, node_dir)

    def _compile_platform(self, base_dir, network_model, renderer):
        platform_compiler = KatharaCompiler()
        platform_data = platform_compiler.compile(network_model)
        platform_renderer = KatharaRenderer(renderer.env)
        rendered = platform_renderer.render(platform_data)
        for entry in rendered:
            write_rendered_to_file(entry, base_dir)



from typing import Optional, List, Dict, Tuple

from pydantic import BaseModel

from autonetkit.network_model.types import DeviceType


class SimplifiedNode(BaseModel):
    id: Optional[str]
    label: Optional[str]
    type: Optional[DeviceType]
    x: Optional[int] = 0
    y: Optional[int] = 0
    asn: Optional[int]
    target: Optional[str]
    data: Optional[Dict] = {}


class SimplifiedTopology(BaseModel):
    nodes: List[SimplifiedNode]
    links: Optional[List[Tuple[str, str]]] = []



import logging

from autonetkit.design.utils import filters
from autonetkit.design.utils.graph_utils import connected_components
from autonetkit.network_model.topology import Topology
from autonetkit.network_model.types import DeviceType

logger = logging.getLogger(__name__)


def check_layer2_conn(topology: Topology) -> bool:
    """

    @param topology:
    @return:
    """
    valid = True

    components = connected_components(topology)
    if len(components) > 1:
        logger.warning("Disconnected network: %s components", len(components))
        valid = False

    routers = filters.routers(topology)

    routers_present = len(routers) > 0

    if routers_present:
        # check that all hosts connect to a router
        hosts = filters.hosts(topology)
        for host in hosts:
            peers = host.peer_nodes()
            if not any(n.type == DeviceType.ROUTER for n in peers):
                logger.warning("Host %s is not connected to a router", host.label)
                valid = False

    return valid



import logging
import typing
from typing import Optional, List, NamedTuple

import networkx as nx

from autonetkit.network_model.types import DeviceType, PortType, NodeId

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from autonetkit.network_model.node import Node
    from autonetkit.network_model.port import Port
    from autonetkit.network_model.link import Link
    from autonetkit.network_model.topology import Topology


class BoundaryData(NamedTuple):
    p_inside: List['Port']
    p_outside: List['Port']
    n_inside: List['Node']
    n_outside: List['Node']
    links: List['Link']


def topology_to_nx_graph(topology: 'Topology', nodes: Optional[List['Node']] = None) -> nx.MultiGraph:
    """

    @param topology:
    @param nodes:
    @return:
    """
    graph = nx.MultiGraph()
    if nodes:
        node_ids = [n.id for n in nodes]
    else:
        node_ids = [n.id for n in topology.nodes()]

    graph.add_nodes_from(node_ids)

    link_pairs = [(l.n1.id, l.n2.id, {"id": l.id})
                  for l in topology.links()
                  if l.n1.id in graph and l.n2.id in graph]

    graph.add_edges_from(link_pairs)

    return graph


def connected_components(topology: 'Topology',
                         nodes: Optional[List['Node']] = None) -> List[List['Node']]:
    """

    @param topology:
    @param nodes:
    @return:
    """
    if nodes is not None and len(nodes) == 0:
        # otherwise networkx returns all components for graph
        # if want all nodes, then use nodes=None
        return []
    graph = topology_to_nx_graph(topology, nodes)

    components = nx.connected_components(graph)
    components = list(components)
    result = []
    for component in components:
        nodes = [topology.get_node_by_id(nid) for nid in component]
        result.append(nodes)

    return result


def boundary_links(topology: 'Topology', nodes: List['Node']) -> List['Link']:
    """

    @param topology:
    @param nodes:
    @return:
    """
    graph = topology_to_nx_graph(topology)
    node_ids = [n.id for n in nodes]

    edges = nx.edge_boundary(graph, node_ids, data=True)
    edge_ids = [data.get("id") for _, _, data in edges]
    links = [topology.get_link_by_id(lid) for lid in edge_ids]
    return links


def boundary_elements(topology: 'Topology', nodes: List['Node']) -> BoundaryData:
    """

    @param topology:
    @param nodes:
    @return:
    """
    links = boundary_links(topology, nodes)

    p_inside = []
    p_outside = []
    n_inside = []
    n_outside = []

    nodes = set(nodes)
    for link in links:
        if link.n1 in nodes:
            n_inside.append(link.n1)
            n_outside.append(link.n2)
            p_inside.append(link.p1)
            p_outside.append(link.p2)

        else:
            n_inside.append(link.n2)
            n_outside.append(link.n1)
            p_inside.append(link.p2)
            p_outside.append(link.p1)

    return BoundaryData(p_inside, p_outside, n_inside, n_outside, links)


def merge_nodes(topology: 'Topology', nodes: List['Node'],
                label: Optional[str] = None) -> 'Node':
    """

    @param topology:
    @param nodes:
    @param label:
    @return:
    """
    # returns the newly created merged node
    boundary_data = boundary_elements(topology, nodes)
    topology.remove_nodes_from(nodes)
    topology.remove_links_from(boundary_data.links)

    if not label:
        label = ""

    virtual_node = topology.create_node(DeviceType.VIRTUAL, label=label)

    for p_outside in boundary_data.p_outside:
        p_inside = virtual_node.create_port(PortType.LOGICAL)
        topology.create_link(p_inside, p_outside)

    return virtual_node


def split_link(topology: 'Topology', link: ['Link'],
               label: Optional[str] = None) -> 'Node':
    """

    @param topology:
    @param link:
    @param label:
    @return:
    """
    if not label:
        label = ""

    virtual_node = topology.create_node(DeviceType.VIRTUAL, label=label)

    p1 = virtual_node.create_port(PortType.LOGICAL)
    topology.create_link(link.p1, p1)
    p2 = virtual_node.create_port(PortType.LOGICAL)
    topology.create_link(link.p2, p2)

    topology.remove_link(link)

    return virtual_node


def explode_node(topology: 'Topology', node: 'Node') -> List['Link']:
    """

    @param topology:
    @param node:
    @return:
    """
    result = []
    peer_ports = node.peer_ports()
    topology.remove_node(node)

    pairs = [(x, y) for x in peer_ports for y in peer_ports
             if x != y]

    for p1, p2 in pairs:
        link = topology.create_link(p1, p2)
        result.append(link)

    return result


def links_contained_by_nodes(topology: 'Topology', nodes: List['Node']) -> List['Link']:
    """

    @param topology:
    @param nodes:
    @return:
    """
    nodes = set(nodes)
    result = []
    for link in topology.links():
        if link.n1 in nodes and link.n2 in nodes:
            result.append(link)

    return result


def wrap_node_ids(topology: 'Topology', nodes: List[NodeId]) -> List['Node']:
    """

    @param topology:
    @param nodes:
    @return:
    """
    return [topology.get_node_by_id(nid) for nid in nodes]


def force_layout(topology: 'Topology', scale: int = 500):
    """

    @param topology:
    @param scale:
    @return:
    """
    # note applies to global layout, but based on connectivity of provided topology
    try:
        import numpy
    except ImportError:
        logger.warning("Layout requires numpy to be installed")
        return
    graph = topology_to_nx_graph(topology)
    print(graph)
    center = [scale / 2, scale / 2]
    pos = nx.spring_layout(graph, center=center, scale=scale)
    for node_id, (x, y) in pos.items():
        node = topology.get_node_by_id(node_id)
        node.set("x", x)
        node.set("y", y)

    normalise_node_locations(topology)


def normalise_node_locations(topology: 'Topology') -> None:
    """

    @param topology:
    """
    x_min = min(n.get("x") for n in topology.nodes())
    y_min = min(n.get("y") for n in topology.nodes())

    x_offset = y_offset = 0
    if x_min < 0:
        x_offset = 1 - x_min
    if y_min < 0:
        y_offset = 1 - y_min

    if x_offset or y_offset:
        for node in topology.nodes():
            node.set("x", node.get("x") + x_offset)
            node.set("y", node.get("y") + y_offset)


from typing import Dict

from autonetkit.compile.node.base import BaseCompiler
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.node import Node


class LinuxHostCompiler(BaseCompiler):
    """

    """

    def __init__(self):
        super().__init__()

        # TODO: assign this at topology load time in pre-process step
        self.lo0_id = "lo:1"

    def compile(self, network_model: NetworkModel, node: Node) -> Dict:
        """

        @param network_model:
        @param node:
        @return:
        """
        # Note: can expand depending on services defined on host, l7 topologies, etc
        return {}



import os

from autonetkit.render.types import RenderedFileEntry


def write_rendered_to_file(entry: RenderedFileEntry, base_dir: str) -> None:
    """

    @param entry:
    @param base_dir:
    """
    if entry.path:
        path = os.path.join(base_dir, entry.path)
    else:
        path = base_dir

    try:
        os.makedirs(path)
    except FileExistsError:
        pass
    filename = os.path.join(path, entry.filename)
    with open(filename, "w") as fh:
        fh.write(entry.body)



from enum import Enum
from typing import NewType

TopologyId = NewType('TopologyId', str)
NodeId = NewType('NodeId', str)
LinkId = NewType('LinkId', str)
PortId = NewType('PortId', str)
PathId = NewType('PathId', str)


class DeviceType(Enum):
    ROUTER = "Router"
    SWITCH = "Switch"
    HOST = "Host"
    VIRTUAL = "Virtual"
    BROADCAST_DOMAIN = "Broadcast Domain"


LAYER3_DEVICES = {DeviceType.ROUTER, DeviceType.HOST}


class PortType(Enum):
    PHYSICAL = "Physical"
    LOGICAL = "Logical"



from typing import Dict, List

from autonetkit.render.node.base import NodeRenderer
from autonetkit.render.types import RenderedFileEntry


class LinuxHostRenderer(NodeRenderer):
    def render(self, data: Dict) -> List[RenderedFileEntry]:
        """

        @param data:
        @return:
        """
        # Note: Can fill in once hosts setup
        result = []
        return result



from typing import Dict, List

from jinja2 import Environment, PackageLoader

from autonetkit.render.exceptions import RendererNotDefined
from autonetkit.render.node.linux_host import LinuxHostRenderer
from autonetkit.render.node.quagga import QuaggaRenderer
from autonetkit.render.types import RenderedFileEntry


class Renderer:
    """

    """

    def __init__(self):
        self.env = Environment(
            loader=PackageLoader('autonetkit', 'render/templates'),
            trim_blocks=True, lstrip_blocks=True
        )

        self.target_map = {
            "quagga": QuaggaRenderer(self.env),
            "linux": LinuxHostRenderer(self.env)
        }

    def render_node(self, compiled_data: Dict, target: str) -> List[RenderedFileEntry]:
        """

        @param compiled_data:
        @param target:
        @return:
        """
        try:
            renderer = self.target_map[target]
        except KeyError:
            raise RendererNotDefined(target)

        return renderer.render(compiled_data)



import itertools
from collections import defaultdict

from autonetkit.design.utils import filters
from autonetkit.load.model import StructuredTopology
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.types import PortType, NodeId, LinkId, PortId, DeviceType


def process_structured_topology(structure_topology: StructuredTopology) -> NetworkModel:
    """

    @param structure_topology:
    @return:
    """
    network_model = NetworkModel()

    t_in = network_model.get_topology("input")
    for node in structure_topology.nodes:
        for port in node.ports:
            if port.type == PortType.PHYSICAL:
                port.label = f"eth{port.slot}"
            if port.loopback_zero:
                port.label = "lo0"

    node_ids = {n.id for n in structure_topology.nodes}
    link_ids = {n.id for n in structure_topology.nodes}
    port_ids = {n.id for n in structure_topology.nodes}

    node_ids_gen = (NodeId(f"n{x}") for x in itertools.count()
                    if x not in node_ids)
    link_ids_gen = (LinkId(f"l{x}") for x in itertools.count()
                    if x not in link_ids)
    port_ids_gen = (PortId(f"p{x}") for x in itertools.count()
                    if x not in port_ids)

    for node in structure_topology.nodes:
        if node.id is None:
            node.id = next(node_ids_gen)
        for port in node.ports:
            if port.id is None:
                port.id = next(port_ids_gen)

    for link in structure_topology.links:
        if link.id is None:
            link.id = next(link_ids_gen)

    apply_defaults(structure_topology)
    vals_to_map = ["x", "y", "asn", "target"]
    for node in structure_topology.nodes:
        t_node = t_in.create_node(node.type, node.label, id=node.id)
        ndict = node.dict()
        for key in vals_to_map:
            t_node.set(key, ndict.get(key))

        # copy in data values
        for key, val in node.data.items():
            t_node.set(key, val)

        for port in node.ports:
            t_port = t_in.create_port(t_node, port.type, port.label, id=port.id)
            t_port.set("slot", port.slot)

            # copy in data values
            for key, val in port.data.items():
                t_port.set(key, val)

            if port.loopback_zero:
                t_node.set("lo0_id", port.id)

    # build map for fast lookup
    nodes_by_label = {}
    ports_by_node_slot = defaultdict(dict)
    for node in t_in.nodes():
        nodes_by_label[node.label] = node
        for port in filters.physical_ports(node):
            ports_by_node_slot[node.id][port.slot] = port

    for link in structure_topology.links:
        n1 = nodes_by_label[link.n1]
        n2 = nodes_by_label[link.n2]
        p1 = ports_by_node_slot[n1.id][link.p1]
        p2 = ports_by_node_slot[n2.id][link.p2]

        t_link = t_in.create_link(p1, p2, id=link.id)
        # copy in data values
        for key, val in link.data.items():
            t_link.set(key, val)

    return network_model


def apply_defaults(topology: StructuredTopology):
    """

    @param topology:
    """
    # TODO: allow passing config to apply
    DEFAULT_ASN = 1
    DEFAULT_OSPF_AREA = 0
    DEFAULT_TARGET = {
        DeviceType.ROUTER: "quagga",
        DeviceType.HOST: "linux",
        DeviceType.SWITCH: "mock_switch",
    }

    for node in topology.nodes:
        device_type = node.type
        if node.asn is None:
            node.asn = DEFAULT_ASN

        if node.target is None:
            node.target = DEFAULT_TARGET[device_type]

        if device_type == DeviceType.ROUTER:
            if node.data.get("ospf_area") is None:
                node.data["ospf_area"] = DEFAULT_OSPF_AREA



from jinja2 import Template

from autonetkit.design.utils import filters
from autonetkit.design.utils.filters import find_node_by_label
from autonetkit.design.utils.general import group_by
from autonetkit.design.utils.graph_utils import topology_to_nx_graph, wrap_node_ids
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.types import DeviceType, PortType
from autonetkit.webserver.publish import publish_model_to_webserver

network_model = NetworkModel()

t_phy = network_model.create_topology("physical")

r1 = t_phy.create_node(DeviceType.ROUTER, "r1")
r1.set("x", 0)
r1.set("y", 0)
r1.set("asn", 1)
r2 = t_phy.create_node(DeviceType.ROUTER, "r2")
r3 = t_phy.create_node(DeviceType.ROUTER, "r3")
r4 = t_phy.create_node(DeviceType.ROUTER, "r4")
r5 = t_phy.create_node(DeviceType.ROUTER, "r5")
h1 = t_phy.create_node(DeviceType.HOST, "h1")
h2 = t_phy.create_node(DeviceType.HOST, "h2")

properties = {
    "r2": (250, 0, 1),
    "r3": (0, 250, 1),
    "r4": (250, 250, 1),
    "r5": (500, 125, 2),
    "h1": (125, 125, 1),
    "h2": (500, 250, 2),
}

for node_id, (x, y, asn) in properties.items():
    node = find_node_by_label(t_phy, node_id)
    node.set("x", x)
    node.set("y", y)
    node.set("asn", asn)

# create ports
r1p1 = r1.create_port(PortType.PHYSICAL)
h1p1 = h1.create_port(PortType.PHYSICAL)
# and link them
t_phy.create_link(r1p1, h1p1)

# or create directly
t_phy.create_link(r1.create_port(PortType.PHYSICAL), r2.create_port(PortType.PHYSICAL))
# or in a loop
pairs = [(r1, r2), (r1, r3), (r2, r4),
         (r3, r4), (r2, r5), (r4, r5), (r5, h2)]
for n1, n2 in pairs:
    t_phy.create_link(n1.create_port(PortType.PHYSICAL), n2.create_port(PortType.PHYSICAL))

# create loopbacks
routers = filters.routers(t_phy)
for node in t_phy.nodes():
    lo0 = node.create_port(PortType.LOGICAL)
    node.set("lo0_id", lo0.id)

# assign port labels
for node in t_phy.nodes():
    physical_ports = filters.physical_ports(node)
    for index, port in enumerate(physical_ports):
        port.set("label", f"eth{index}")

t_ip = network_model.create_topology("ip")
t_ip.add_nodes_from(t_phy.nodes())
t_ip.add_links_from(t_phy.links())
grouped = group_by(t_ip.nodes(), "asn")
for asn, nodes in grouped.items():
    for index, node in enumerate(nodes):
        lo0 = node.loopback_zero()
        loopback_ip = f"172.16.{asn}.{index}"
        lo0.set("ip", loopback_ip)

    links = [l for l in t_ip.links()
             if l.n1.get("asn") == l.n2.get("asn") == asn]
    for index, link in enumerate(links):
        prefix = f"10.{asn}.{index}"
        network = prefix + ".0"
        link.p1.set("ip", prefix + ".1")
        link.p1.set("network", network)
        link.p2.set("ip", prefix + ".2")
        link.p2.set("network", network)

# inter-as links
links = [l for l in t_ip.links()
         if l.n1.get("asn") != l.n2.get("asn")]
for index, link in enumerate(links):
    prefix = f"10.0.{index}"
    network = prefix + ".0"
    link.p1.set("ip", prefix + ".1")
    link.p1.set("network", network)
    link.p2.set("ip", prefix + ".2")

t_ospf = network_model.create_topology("ospf")
t_ospf.add_nodes_from(routers)
ebgp_links = [l for l in t_phy.links()
              if l.n1.get("asn") == l.n2.get("asn")]
t_ospf.add_links_from(ebgp_links)

t_ibgp = network_model.create_topology("ibgp")
t_ibgp.add_nodes_from(routers)
ibgp_pairs = [(n1, n2) for n1 in t_ibgp.nodes()
              for n2 in t_ibgp.nodes()
              if n1 != n2 and n1.get("asn") == n2.get("asn")]
for n1, n2 in ibgp_pairs:
    p1 = n1.loopback_zero()
    p2 = n2.loopback_zero()
    t_ibgp.create_link(p1, p2)

t_ebgp = network_model.create_topology("ebgp")
t_ebgp.add_nodes_from(routers)
ebgp_links = [l for l in t_phy.links()
              if l.n1.get("asn") != l.n2.get("asn")]
t_ebgp.add_links_from(ebgp_links)

# analysis
import networkx as nx

graph = topology_to_nx_graph(t_phy)
path = nx.shortest_path(graph, h1.id, h2.id)
path = wrap_node_ids(t_phy, path)

p1 = t_phy.create_node_path(path)

# Compile device models
compiled = {}

for node in filters.routers(t_phy):
    data = {
        "hostname": node.label,
        "interfaces": [],
        "asn": node.get("asn")
    }
    for port in filters.physical_ports(node):
        ip_port = t_ip.get_port_by_id(port.id)
        data["interfaces"].append({
            "id": port.label,
            "ip": ip_port.get("ip")
        })

    ospf_node = t_ospf.get_node_by_id(node.id)
    ospf_enabled = ospf_node.degree() > 0
    data["ospf"] = {"networks": [],
                    "enabled":ospf_enabled}

    for port in filters.physical_ports(ospf_node):
        if not port.connected:
            continue
        ip_port = t_ip.get_port_by_id(port.id)
        network = ip_port.get("network")
        data["ospf"]["networks"].append(network)


    ebgp_node = t_ebgp.get_node_by_id(node.id)
    data["ebgp"] = {"neighbors": []}
    for peer in ebgp_node.peer_nodes():
        ip_peer = t_ip.get_node_by_id(peer.id)
        peer_ip = ip_peer.loopback_zero().get("ip")
        data["ebgp"]["neighbors"].append({
            "ip": peer_ip,
            "asn": peer.get("asn")
        })

    ibgp_node = t_ibgp.get_node_by_id(node.id)
    bgp_enabled = ebgp_node.degree() > 0 or ibgp_node.degree() > 0
    data["bgp_enabled"] = bgp_enabled
    data["ibgp"] = {"neighbors": []}
    for peer in ibgp_node.peer_nodes():
        ip_peer = t_ip.get_node_by_id(peer.id)
        peer_ip = ip_peer.loopback_zero().get("ip")
        data["ibgp"]["neighbors"].append({
            "ip": peer_ip,
            "asn": peer.get("asn")
        })

    compiled[node] = data

for node in filters.hosts(t_phy):
    data = {
        "hostname": node.label,
        "interfaces": []
    }

    for port in filters.physical_ports(node):
        ip_port = t_ip.get_port_by_id(port.id)
        data["interfaces"].append({
            "id": port.label,
            "ip": ip_port.get("ip")
        })

    compiled[node] = data

# and render using template
rtr_template_str = """
! router
hostname {{ data.hostname }}
{% for interface in data.interfaces %}
{{interface.id}} {{ interface.ip}} up
{% endfor %}
{% if data.ospf.enabled %}
!
router ospf
{% for network in data.ospf.networks %}
    network {{network}}
{% endfor %}
!
{% endif %}
{% if data.bgp_enabled %}
router bgp {{ asn }}
{% for peer in data.ebgp.neighbors %}
    neighbor {{peer.ip}} {{peer.asn}}
{% endfor %}
{% for peer in data.ibgp.neighbors %}
    neighbor {{peer.ip}} {{peer.asn}}
{% endfor %}
{% endif %}
!
"""
host_template_str = """
! host
hostname {{ data.hostname }}
{% for interface in data.interfaces %}
{{interface.id}} {{ interface.ip}} up
{% endfor %}
"""

templates = {
    DeviceType.ROUTER: Template(rtr_template_str, trim_blocks=True),
    DeviceType.HOST: Template(host_template_str, trim_blocks=True)
}

for node, data in compiled.items():
    template = templates[node.type]
    rendered = template.render(data=data)
    print(rendered)

publish_model_to_webserver(network_model)



from autonetkit.design.builder import Builder
from autonetkit.load.load_graphml import import_from_graphml
from autonetkit.load.validation import validate
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.packager.mapping import get_platform_builder
from autonetkit.webserver.publish import publish_model_to_webserver


class BaseWorkflow:
    """

    """

    def __init__(self):
        pass

    def run(self, network_model: NetworkModel, target_platform: str):
        """

        @param network_model:
        @param target_platform:
        """
        t_in = network_model.get_topology("input")
        validate(t_in)
        self.build(network_model)
        self.render(network_model, target_platform)
        self.publish(network_model)

    @staticmethod
    def load(input_file):
        """

        @param input_file:
        @return:
        """
        network_model = import_from_graphml(input_file)
        return network_model

    @staticmethod
    def publish(network_model):
        """

        @param network_model:
        """
        host = "http://0.0.0.0:8080"
        publish_model_to_webserver(network_model, host)

    @staticmethod
    def build(network_model):
        """

        @param network_model:
        """
        builder = Builder(network_model)
        builder.build()

    @staticmethod
    def render(network_model, target_platform):
        """

        @param network_model:
        @param target_platform:
        """
        builder = get_platform_builder(target_platform)
        builder.build(network_model)



from typing import Dict, List

from autonetkit.render.node.base import NodeRenderer
from autonetkit.render.types import RenderedFileEntry


class QuaggaRenderer(NodeRenderer):
    def render(self, data: Dict) -> List[RenderedFileEntry]:
        """

        @param data:
        @return:
        """
        result = []
        quagga_path = "etc/quagga"

        zebra_template = self.env.get_template("quagga/zebra.conf")
        result.append(RenderedFileEntry(
            zebra_template.render(node=data),
            quagga_path,
            "zebra.conf"))

        ospfd_template = self.env.get_template("quagga/ospfd.conf")
        result.append(RenderedFileEntry(
            ospfd_template.render(node=data),
            quagga_path,
            "ospfd.conf"))

        bgpd_template = self.env.get_template("quagga/bgpd.conf")
        result.append(RenderedFileEntry(
            bgpd_template.render(node=data),
            quagga_path,
            "bgpd.conf"))

        daemons_template = self.env.get_template("quagga/daemons.conf")
        result.append(RenderedFileEntry(
            daemons_template.render(node=data),
            quagga_path,
            "daemons.conf"))

        return result



import ipaddress
import json
import logging
from enum import Enum

import requests

logger = logging.getLogger(__name__)


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        """

        @param obj:
        @return:
        """
        if isinstance(obj, Enum):
            return obj.name
        elif isinstance(obj, ipaddress.IPv4Network):
            return str(obj)
        elif isinstance(obj, ipaddress.IPv4Address):
            return str(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def publish_model_to_webserver(network_model, host="http://0.0.0.0:8080"):
    """

    @param network_model:
    @param host:
    @return:
    """
    exported = network_model.export()
    exported_json = json.dumps(exported, cls=ComplexEncoder, indent=4)
    try:
        result = requests.post(f"{host}/data", data=exported_json)
    except requests.exceptions.ConnectionError:
        logger.warning("Unable to post data to server: %s", host)
        return

    try:
        result.raise_for_status()
    except requests.exceptions.HTTPError as err:
        logger.warning(err)
    else:
        logger.info("Published to webserver")



from typing import Dict, List

from autonetkit.render.platform.base import PlatformRenderer
from autonetkit.render.types import RenderedFileEntry


class KatharaRenderer(PlatformRenderer):
    def render(self, data: Dict) -> List[RenderedFileEntry]:
        """

        @param data:
        @return:
        """
        result = []

        lab_template = self.env.get_template("kathara/lab.conf")
        result.append(RenderedFileEntry(
            lab_template.render(data=data["lab"]),
            None,
            "lab.conf"))

        node_startup_template = self.env.get_template("kathara/node.startup")
        for node in data["nodes"]:
            filename = f"{node['label']}.startup"
            result.append(RenderedFileEntry(
                node_startup_template.render(data=node),
                None,
                filename))

        return result

    def render_node(self, node_data: Dict) -> List[RenderedFileEntry]:
        """

        @param node_data:
        @return:
        """
        result = []

        hostname_template = self.env.get_template("kathara/hostname")
        result.append(RenderedFileEntry(
            hostname_template.render(data=node_data["kathara"]["hostname"]),
            "etc",
            "hostname"))

        shadow_template = self.env.get_template("kathara/shadow")
        result.append(RenderedFileEntry(
            shadow_template.render(),
            "etc",
            "shadow"))

        ssh_template = self.env.get_template("kathara/ssh")
        result.append(RenderedFileEntry(
            ssh_template.render(),
            "etc/ssh",
            "sshd_config"))

        return result



from autonetkit.design.utils import filters
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.topology import Topology
from autonetkit.network_model.types import DeviceType


def _build_igp_base(network_model: NetworkModel, topology: Topology):
    t_l2_conn = network_model.get_topology("layer2_conn")
    routers = filters.routers(t_l2_conn)
    topology.add_nodes_from(routers)
    links = [l for l in t_l2_conn.links()
             if l.n1.type == l.n2.type == DeviceType.ROUTER
             and l.n1.get("asn") == l.n2.get("asn")]
    topology.add_links_from(links)
    # remove the zero degree routers -> no IGP to configure
    single_routers = [n for n in topology.nodes() if n.degree() == 0]
    topology.remove_nodes_from(single_routers)


def build_ospf(network_model: NetworkModel):
    """

    @param network_model:
    """
    t_ospf = network_model.create_topology("ospf")
    t_in = network_model.get_topology("input")
    _build_igp_base(network_model, t_ospf)
    for node in t_ospf.nodes():
        input_node = t_in.get_node_by_id(node.id)
        # TODO: allow specifying the OSPF area on input topology
        node.set("area", input_node.get("ospf_area"))
        node_area = node.get("area")
        for port in filters.physical_ports(node):
            port.set("area", node_area)


def build_isis(network_model: NetworkModel):
    """

    @param network_model:
    """
    t_isis = network_model.create_topology("isis")
    _build_igp_base(network_model, t_isis)


def build_ibgp(network_model: NetworkModel):
    """

    @param network_model:
    """
    t_ibgp = network_model.create_topology("ibgp")
    t_l2_conn = network_model.get_topology("layer2_conn")
    routers = filters.routers(t_l2_conn)
    t_ibgp.add_nodes_from(routers)
    pairs = [(s, t)
             for s in t_ibgp.nodes()
             for t in t_ibgp.nodes()
             if s.get("asn") == t.get("asn")
             if s != t]

    # get loopback

    for n1, n2 in pairs:
        p1 = n1.loopback_zero()
        p2 = n2.loopback_zero()
        t_ibgp.create_link(p1, p2)


def build_ebgp(network_model: NetworkModel):
    """

    @param network_model:
    """
    t_ebgp = network_model.create_topology("ebgp")
    t_l2_conn = network_model.get_topology("layer2_conn")
    routers = filters.routers(t_l2_conn)
    t_ebgp.add_nodes_from(routers)

    ebgp_links = [l for l in t_l2_conn.links()
                  if l.n1.get("asn") != l.n2.get("asn")]
    t_ebgp.add_links_from(ebgp_links)



from autonetkit.design.utils.graph_utils import force_layout
from autonetkit.load.common import build_model_from_nx_graph
from autonetkit.workflow.workflow import BaseWorkflow


def main():
    """

    """
    workflow = BaseWorkflow()

    import networkx as nx
    graph = nx.random_internet_as_graph(100)
    network_model = build_model_from_nx_graph(graph)
    t_in = network_model.get_topology("input")

    asn = 1
    for node_id, data in graph.nodes(data=True):
        node = t_in.get_node_by_id(str(node_id))
        node.set("role", data["type"])
        node.set("asn", asn)
        asn += 1

    t_in = network_model.get_topology("input")
    force_layout(t_in, scale=500)
    workflow.run(network_model, target_platform="kathara")


if __name__ == '__main__':
    main()



import typing
from typing import List

from autonetkit.network_model.exceptions import NodeNotFound
from autonetkit.network_model.types import DeviceType, PortType

if typing.TYPE_CHECKING:
    from autonetkit.network_model.port import Port
    from autonetkit.network_model.node import Node
    from autonetkit.network_model.topology import Topology


def filter_nodes(topology: 'Topology', **kwargs) -> List['Node']:
    """

    @param topology:
    @param kwargs:
    @return:
    """
    result = []
    for node in topology.nodes():
        matched = all(node.get(key) == val
                      for key, val in kwargs.items())
        if matched:
            result.append(node)

    return result


def routers(topology: 'Topology') -> List['Node']:
    """

    @param topology:
    @return:
    """
    return filter_nodes(topology, type=DeviceType.ROUTER)


def switches(topology: 'Topology') -> List['Node']:
    """

    @param topology:
    @return:
    """
    return filter_nodes(topology, type=DeviceType.SWITCH)


def hosts(topology: 'Topology') -> List['Node']:
    """

    @param topology:
    @return:
    """
    return filter_nodes(topology, type=DeviceType.HOST)


def broadcast_domains(topology) -> List['Node']:
    """

    @param topology:
    @return:
    """
    return filter_nodes(topology, type=DeviceType.BROADCAST_DOMAIN)


def physical_ports(node: 'Node') -> List['Port']:
    """

    @param node:
    @return:
    """
    return [p for p in node.ports()
            if p.type == PortType.PHYSICAL]


def find_node_by_label(topology: 'Topology', label) -> 'Node':
    """

    @param topology:
    @param label:
    @return:
    """
    for node in topology.nodes():
        if node.label == label:
            return node

    raise NodeNotFound()



import itertools
from collections import defaultdict
from typing import Dict

from autonetkit.network_model.exceptions import TopologyNotFound
from autonetkit.network_model.topology import Topology
from autonetkit.network_model.types import TopologyId, NodeId, PortId, LinkId, PathId


class NetworkModel:
    """

    """

    def __init__(self):
        self._topologies: Dict[TopologyId, Topology] = {}

        # create mandatory topologies
        self.create_topology("input")
        self.create_topology("physical")

        self.node_globals: Dict[NodeId, Dict] = defaultdict(dict)
        self.port_globals: Dict[PortId, Dict] = defaultdict(dict)

        self.node_global_keys = {"type", "label", "asn", "x", "y", "target", "lo0_id"}
        self.port_global_keys = {"type", "label", "slot"}

        self.used_node_ids = set()
        self.used_link_ids = set()
        self.used_port_ids = set()
        self.used_path_ids = set()

    @property
    def _node_ids(self):
        for x in itertools.count():
            candidate = f"n{x}"
            if candidate not in self.used_node_ids:
                self.used_node_ids.add(candidate)
                yield candidate

    @property
    def _link_ids(self):
        for x in itertools.count():
            candidate = f"l{x}"
            if candidate not in self.used_link_ids:
                self.used_link_ids.add(candidate)
                yield candidate

    @property
    def _port_ids(self):
        for x in itertools.count():
            candidate = f"p{x}"
            if candidate not in self.used_port_ids:
                self.used_port_ids.add(candidate)
                yield candidate

    @property
    def _path_ids(self):
        for x in itertools.count():
            candidate = f"pa{x}"
            if candidate not in self.used_path_ids:
                self.used_path_ids.add(candidate)
                yield candidate

    def get_topology(self, name) -> Topology:
        """

        @param name:
        @return:
        """
        try:
            return self._topologies[name]
        except KeyError:
            raise TopologyNotFound(name)

    def create_topology(self, name) -> Topology:
        """

        @param name:
        @return:
        """
        topology = Topology(self, name)
        self._topologies[name] = topology
        return topology

    def generate_node_id(self) -> NodeId:
        """

        @return:
        """
        return NodeId(next(self._node_ids))

    def generate_link_id(self) -> LinkId:
        """

        @return:
        """
        return LinkId(next(self._link_ids))

    def generate_port_id(self) -> PortId:
        """

        @return:
        """
        return PortId(next(self._port_ids))

    def generate_path_id(self) -> PathId:
        """

        @return:
        """
        return PathId(next(self._path_ids))

    def export(self) -> Dict:
        """

        @return:
        """
        result = {}
        for name, topology in self._topologies.items():
            result[name] = topology.export()

        return result



from collections import Counter
from enum import Enum

from autonetkit.design.utils import filters
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.topology import Topology
from autonetkit.network_model.types import DeviceType


def build_ip_base(network_model: NetworkModel, topology: Topology):
    """

    @param network_model:
    @param topology:
    """
    t_l2 = network_model.get_topology("layer2")
    topology.add_nodes_from(t_l2.nodes())
    topology.add_links_from(t_l2.links())

    bc_nodes = filters.broadcast_domains(topology)

    for node in bc_nodes:
        peers = node.peer_nodes()
        peer_asns = {n.get("asn") for n in peers}
        peer_types = {n.type for n in peers}
        if len(peer_asns) == 1:
            if peer_types == {DeviceType.ROUTER}:
                node.set("bc_type", IpBcTypes.BACKBONE)
            else:
                node.set("bc_type", IpBcTypes.HOSTS)

        else:
            node.set("bc_type", IpBcTypes.INTER_DOMAIN)


def allocate_asns(topology: Topology):
    """

    @param topology:
    """
    bc_nodes = filters.broadcast_domains(topology)

    for node in bc_nodes:
        peer_asns = sorted(n.get("asn") for n in node.peer_nodes())
        if len(set(peer_asns)) == 1:
            asn = peer_asns[0]
        else:
            counted = Counter(peer_asns)
            asn = counted.most_common(1)[0][0]
        node.set("asn", asn)


def map_blocks_to_ports(topology):
    """

    @param topology:
    """
    bc_nodes = filters.broadcast_domains(topology)
    for bc_node in bc_nodes:
        subnet = bc_node.get("network")
        hosts = subnet.hosts()
        for peer_port in bc_node.peer_ports():
            peer_port.set("network", subnet)
            host_ip = next(hosts)
            peer_port.set("ip", host_ip)


class IpBcTypes(Enum):
    HOSTS = 1
    BACKBONE = 2
    INTER_DOMAIN = 3



import typing
from typing import Dict

from autonetkit.compile.node.linux_host import LinuxHostCompiler
from autonetkit.compile.node.quagga import QuaggaCompiler

if typing.TYPE_CHECKING:
    from autonetkit.network_model.network_model import NetworkModel
    from autonetkit.network_model.node import Node


class TargetCompilerNotFound(Exception):
    pass


class NodeCompiler:
    """

    """

    def __init__(self):
        self.node_mapping = {
            "quagga": QuaggaCompiler(),
            "linux": LinuxHostCompiler()
        }

    def compile_node(self, network_model: 'NetworkModel', node: 'Node') -> Dict:
        """

        @param network_model:
        @param node:
        @return:
        """
        target = node.get("target")
        try:
            compiler = self.node_mapping[target]
        except KeyError:
            raise TargetCompilerNotFound(target)

        return compiler.compile(network_model, node)



from collections import defaultdict

import networkx as nx

from autonetkit.load.common import add_loopback
from autonetkit.design.utils.graph_utils import normalise_node_locations
from autonetkit.load.model import StructuredNode, StructuredPort, StructuredTopology, StructuredLink
from autonetkit.load.preprocess import process_structured_topology
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.types import PortType, DeviceType


def import_from_graphml(filename: str) -> NetworkModel:
    """

    @param filename:
    @return:
    """
    with open(filename) as fh:
        graph = nx.read_graphml(fh)

    node_map = {}
    node_port_map = defaultdict(dict)

    topology = StructuredTopology()

    reserved_node_keys = {"x", "y", "device_type", "label" "asn", "target"}
    reserved_edge_keys = {}

    for nx_node_id, node_data in graph.nodes(data=True):
        label = node_data.get("label").strip()

        node_metadata = {}

        for key, val in node_data.items():
            if key not in reserved_node_keys:
                node_metadata[key] = val

        # create loopback zero
        device_type = node_data.get("device_type").title()
        get = node_data.get("x")
        node = StructuredNode(type=device_type,
                              label=label,
                              x=get,
                              y=node_data.get("y"),
                              asn=node_data.get("asn"),
                              target=node_data.get("target"),
                              data = node_metadata
                              )

        topology.nodes.append(node)
        node_map[nx_node_id] = node

        if node.type in {DeviceType.ROUTER, DeviceType.HOST}:
            add_loopback(node)

        edges = [data for _, _, data in graph.in_edges(nx_node_id, data=True)]
        edges += [data for _, _, data in graph.out_edges(nx_node_id, data=True)]
        for slot, edge_data in enumerate(edges):
            port = StructuredPort(type=PortType.PHYSICAL, slot=slot)
            node.ports.append(port)
            edge_id = edge_data["id"]
            node_port_map[nx_node_id][edge_id] = port

    for src, dst, edge_data in graph.edges(data=True):
        edge_id = edge_data["id"]
        n1 = node_map[src].label
        n2 = node_map[dst].label
        p1 = node_port_map[src][edge_id].slot
        p2 = node_port_map[dst][edge_id].slot

        edge_metadata = {}

        for key, val in edge_data.items():
            if key not in reserved_edge_keys:
                edge_metadata[key] = val

        link = StructuredLink(n1=n1, n2=n2, p1=p1, p2=p2, data=edge_metadata)
        topology.links.append(link)

    # TODO: later match this to the YAML physical inventory etc
    network_model = process_structured_topology(topology)
    t_in = network_model.get_topology("input")

    normalise_node_locations(t_in)

    # and reset network model ids in use

    return network_model



import abc
from typing import Dict

from autonetkit.network_model.network_model import NetworkModel


class BasePlatformCompiler:
    """

    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def compile(self, network_model: NetworkModel) -> Dict:
        """

        @param network_model:
        """
        pass



class NodeNotFound(Exception):
    pass


class PortNotFound(Exception):
    pass


class LinkNotFound(Exception):
    pass


class TopologyNotFound(Exception):
    pass


class PortIdInUse(Exception):
    pass


class LinkIdInUse(Exception):
    pass


class PathIdInUse(Exception):
    pass


class NodeIdInUse(Exception):
    pass


class NumericNodeId(Exception):
    pass



import typing
from collections import defaultdict

import networkx as nx

from autonetkit.design.utils import filters
from autonetkit.load.model import StructuredTopology, StructuredNode, StructuredPort, StructuredLink
from autonetkit.load.preprocess import process_structured_topology
from autonetkit.load.simplified import SimplifiedTopology
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.topology import Topology
from autonetkit.network_model.types import DeviceType, PortType, LAYER3_DEVICES


def label_physical_ports(topology: Topology, port_prefix) -> None:
    """

    @param topology:
    @param port_prefix:
    """
    # TODO: later can skip this step if assigned by user
    for node in topology.nodes():
        physical_ports = filters.physical_ports(node)
        for index, port in enumerate(physical_ports):
            port.set("label", f"{port_prefix}{index}")
            port.set("slot", index)


def build_model_from_nx_graph(graph: typing.Union[nx.Graph, nx.DiGraph]) -> NetworkModel:
    """

    @param graph:
    @return: 
    """
    raw_topology = {"nodes": [], "links": []}
    for nid in graph.nodes():
        raw_topology["nodes"].append({
            "id": nid,
            "label": f"r{nid}"
        })

    for src, dst in graph.edges():
        pair = (f"r{src}", f"r{dst}")
        raw_topology["links"].append(pair)

    simplified_topology = SimplifiedTopology(**raw_topology)
    topology = transform_simplified_to_structured_topology(simplified_topology)
    network_model = process_structured_topology(topology)

    return network_model


def transform_simplified_to_structured_topology(simplified_topology):
    """

    @param simplified_topology: 
    @return: 
    """
    for node in simplified_topology.nodes:
        if node.type is None:
            node.type = DeviceType.ROUTER

    topology = StructuredTopology()
    node_map = {}
    port_map = defaultdict(int)
    for simplified_node in simplified_topology.nodes:
        structured_node = StructuredNode(**simplified_node.dict())
        node_map[simplified_node.label] = structured_node
        if structured_node.type in LAYER3_DEVICES:
            port = StructuredPort(type=PortType.LOGICAL, loopback_zero=True)
            structured_node.ports.append(port)

        topology.nodes.append(structured_node)
    for simplified_link in simplified_topology.links:
        n1_label = simplified_link[0]
        n2_label = simplified_link[1]
        n1 = node_map[n1_label]
        n2 = node_map[n2_label]
        slot1 = port_map[n1_label]
        slot2 = port_map[n2_label]
        port_map[n1_label] += 1
        port_map[n2_label] += 2

        p1 = StructuredPort(type=PortType.PHYSICAL, slot=slot1)
        n1.ports.append(p1)
        p2 = StructuredPort(type=PortType.PHYSICAL, slot=slot2)
        n2.ports.append(p2)

        link = StructuredLink(n1=n1_label, n2=n2_label, p1=slot1, p2=slot2)
        topology.links.append(link)
    return topology


def add_loopback(node):
    """

    @param node: 
    """
    port = StructuredPort(type=PortType.LOGICAL, loopback_zero=True)
    node.ports.append(port)



import abc
from typing import Dict

from autonetkit.network_model.network_model import NetworkModel


class AbstractPlatformPackager:
    """

    """

    def __init__(self, output_dir="output"):
        self.output_dir = output_dir

    @abc.abstractmethod
    def build(self, network_model: NetworkModel) -> Dict:
        """

        @param network_model:
        """
        pass



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name="autonetkit",
    version="0.12.2",
    description='Automatic configuration generation',
    long_description='Automatic configuration generation',

    entry_points={
        'console_scripts': [
            'autonetkit = autonetkit.console:main',
            'webserver = autonetkit.webserver:main',
        ],
    },

    author='Simon Knight',
    author_email="simon.knight@gmail.com",
    url="http://www.autonetkit.org",

    packages=find_packages(exclude=('tests', 'docs')),

    include_package_data=True,  # include data from MANIFEST.in

    download_url=("http://pypi.python.org/pypi/autonetkit"),

    install_requires=[
        'netaddr',
        'networkx',
        'Jinja2',
        'aiohttp',
        'requests',
        'pydantic'
    ],

    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Networking",
        "Topic :: System :: Software Distribution",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],

)



import abc
from typing import List

from jinja2 import Environment

from autonetkit.network_model.node import Node
from autonetkit.render.types import RenderedFileEntry


class PlatformRenderer:
    """

    """

    def __init__(self, env: Environment):
        self.env: Environment = env

    @abc.abstractmethod
    def render(self, node: Node) -> List[RenderedFileEntry]:
        """

        @param node:
        """
        pass



import logging
from typing import Dict, List

from autonetkit.compile.platform.base import BasePlatformCompiler
from autonetkit.design.utils import filters
from autonetkit.network_model.exceptions import PortNotFound
from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.node import Node
from autonetkit.network_model.types import DeviceType

logger = logging.getLogger(__name__)


class KatharaCompiler(BasePlatformCompiler):
    def compile(self, network_model: NetworkModel) -> Dict:
        """

        @param network_model:
        @return:
        """
        result = {}
        t_phy = network_model.get_topology("physical")

        result["nodes"] = self._compile_node_startup(network_model, t_phy)
        result["lab"] = self._compile_lab(t_phy)

        return result

    def _compile_lab(self, t_phy) -> Dict:
        result = {
            "nodes": []
        }

        for node in t_phy.nodes():
            interfaces = []
            for port in filters.physical_ports(node):
                if port.degree() > 0:
                    if port.degree() > 1:
                        logger.warning("Port %s on %s has multiple physical links: only using first",
                                       port.label, node.label)
                    # TODO: note this assumes single physical link per port
                    link = port.links()[0]
                    link_id = link.id

                    interfaces.append({
                        "slot": port.slot,
                        "link_id": link_id
                    })

            entry = {
                "label": node.label,
                "interfaces": interfaces
            }
            result["nodes"].append(entry)

        return result

    def _compile_node_startup(self, network_model, t_phy) -> List:
        result = []
        t_ipv4 = network_model.get_topology("ipv4")
        for node in t_phy.nodes():
            interfaces = []
            for port in filters.physical_ports(node):
                data = {
                    "label": port.label,
                    "connected": port.connected
                }

                if port.connected:
                    try:
                        ipv4_port = t_ipv4.get_port_by_id(port.id)
                    except PortNotFound:
                        # no ip for this port
                        continue

                    data["ip"] = ipv4_port.get("ip")
                    data["netmask"] = ipv4_port.get("network").netmask

                interfaces.append(data)

            services = {}
            if node.type == DeviceType.ROUTER:
                services["zebra"] = True

            result.append({
                "label": node.label,
                "interfaces": interfaces,
                "services": services
            })

        return result

    def compile_node(self, node: Node) -> Dict:
        """

        @param node:
        @return:
        """
        return {
            "hostname": node.label
        }



import typing

from autonetkit.network_model.exceptions import PortNotFound
from autonetkit.network_model.types import NodeId, DeviceType, PortType

if typing.TYPE_CHECKING:
    from autonetkit.network_model.topology import Topology
    from autonetkit.network_model.port import Port
    from autonetkit.network_model.link import Link


class Node:
    """

    """

    def __init__(self, topology: 'Topology', id):
        self.topology: Topology = topology
        self.id: NodeId = id
        self._data = {}

    @property
    def local_data(self) -> typing.Dict:
        """

        @return:
        """
        return self._data

    @property
    def label(self) -> str:
        """

        @return:
        """
        return self.get("label")

    @property
    def type(self) -> DeviceType:
        """

        @return:
        """
        return self.get("type")

    @property
    def global_data(self) -> typing.Dict:
        """

        @return:
        """
        return self.topology.network_model.node_globals[self.id]

    def __repr__(self):
        return f"N {self.get('label')}"

    def ports(self) -> typing.List['Port']:
        """

        @return:
        """
        return self.topology.ports(node=self)

    def links(self) -> typing.List['Link']:
        """

        @return:
        """
        return self.topology.links(node=self)

    def create_port(self, type: PortType) -> 'Port':
        """

        @param type:
        @return:
        """
        port = self.topology.create_port(self, type)
        return port

    def set(self, key, val) -> None:
        """

        @param key:
        @param val:
        """
        if key in self.topology.network_model.node_global_keys:
            self.global_data[key] = val
        else:
            self._data[key] = val

    def get(self, key, default=None):
        """

        @param key:
        @param default:
        @return:
        """
        try:
            if key in self.topology.network_model.node_global_keys:
                return self.global_data[key]
            else:
                return self._data[key]
        except KeyError:
            return default

    def export(self) -> typing.Dict:
        """

        @return:
        """
        data = self.global_data.copy()
        data.update(self._data.copy())
        return data

    def peer_nodes(self) -> typing.List['Node']:
        """

        @return:
        """
        result = []
        for link in self.links():
            other = link.other_node(self)
            result.append(other)

        return result

    def peer_ports(self) -> typing.List['Port']:
        """

        @return:
        """
        result = []
        for link in self.links():
            if link.n1 == self:
                result.append(link.p2)
            else:
                result.append(link.p1)

        return result

    def degree(self) -> int:
        """

        @return:
        """
        return len(self.links())

    def loopback_zero(self) -> 'Port':
        """

        @return:
        """
        lo0_id = self.get("lo0_id")
        if lo0_id is None:
            raise PortNotFound("loopback_zero")
        else:
            return self.topology.get_port_by_id(lo0_id)



import typing

from autonetkit.network_model.types import PortType, PortId

if typing.TYPE_CHECKING:
    from autonetkit.network_model.link import Link
    from autonetkit.network_model.node import Node


class Port:
    """

    """

    def __init__(self, node: 'Node', id):
        self._node: Node = node
        self.id: PortId = id
        self._data = {}

    def __repr__(self):
        return f"P {self.get('label')}"

    @property
    def local_data(self) -> typing.Dict:
        """

        @return:
        """
        return self._data

    @property
    def label(self) -> str:
        """

        @return:
        """
        return self.get("label")

    @property
    def type(self) -> PortType:
        """

        @return:
        """
        return self.get("type")

    @property
    def slot(self) -> int:
        """

        @return:
        """
        return self.get("slot")

    @property
    def global_data(self) -> typing.Dict:
        """

        @return:
        """
        return self._node.topology.network_model.port_globals[self.id]

    @property
    def node(self) -> 'Node':
        """

        @return:
        """
        return self._node

    def set(self, key, val) -> None:
        """

        @param key:
        @param val:
        """
        if key in self.node.topology.network_model.port_global_keys:
            self.global_data[key] = val
        else:
            self._data[key] = val

    def get(self, key, default=None):
        """

        @param key:
        @param default:
        @return:
        """
        try:
            if key in self.node.topology.network_model.port_global_keys:
                return self.global_data[key]
            else:
                return self._data[key]
        except KeyError:
            return default

    def export(self) -> typing.Dict:
        """

        @return:
        """
        data = self.global_data.copy()
        data.update(self._data.copy())

        data["node"] = self._node.id

        return data

    def links(self) -> typing.List['Link']:
        """

        @return:
        """
        result = [l for l in (self.node.links())
                  if l.p1 == self
                  or l.p2 == self]
        return result

    def peer_nodes(self, unique=True) -> typing.List['Node']:
        """

        @param unique:
        @return:
        """
        result = []
        for link in self.links():
            if link.p1 == self:
                result.append(link.n2)
            else:
                result.append(link.n1)

        if unique:
            result = list(set(result))

        return result

    def peer_ports(self, unique=True) -> typing.List['Port']:
        """

        @param unique:
        @return:
        """
        result = []
        for link in self.links():
            if link.p1 == self:
                result.append(link.p2)
            else:
                result.append(link.p1)

        if unique:
            result = list(set(result))

        return result

    def degree(self) -> int:
        """

        @return:
        """
        return len(self.links())

    @property
    def connected(self) -> bool:
        """

        @return:
        """
        return self.degree() > 0



from typing import List, Optional, Dict

from pydantic import BaseModel

from autonetkit.network_model.types import DeviceType, PortType, LinkId, PortId, NodeId


class StructuredPort(BaseModel):
    id: Optional[PortId]
    slot: Optional[int]
    type: PortType
    label: Optional[str]
    data: Optional[Dict] = {}
    loopback_zero: Optional[bool]


class StructuredNode(BaseModel):
    id: Optional[NodeId]
    type: Optional[DeviceType]
    label: str
    x: Optional[float]
    y: Optional[float]
    asn: Optional[int]
    target: Optional[int]
    loopback_zero_id: Optional[StructuredPort]
    data: Optional[Dict] = {}
    ports: List[StructuredPort] = []


class StructuredLink(BaseModel):
    id: Optional[LinkId]
    n1: str
    n2: str
    p1: int
    p2: int
    data: Optional[Dict] = {}


class StructuredTopology(BaseModel):
    nodes: List[StructuredNode] = []
    links: List[StructuredLink] = []



import abc
from typing import Dict

from autonetkit.network_model.network_model import NetworkModel
from autonetkit.network_model.node import Node


class BaseCompiler:
    """

    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def compile(self, network_model: NetworkModel, node: Node) -> Dict:
        """

        @param network_model:
        @param node:
        """
        pass



import typing

if typing.TYPE_CHECKING:
    from autonetkit.network_model.port import Port
    from autonetkit.network_model.node import Node


class NodePath:
    """

    """

    def __init__(self, topology, id, nodes: typing.List['Node']):
        self.id = id
        self.topology = topology
        self.nodes: typing.List['Node'] = nodes
        self._data = {}

    @property
    def local_data(self) -> typing.Dict:
        """

        @return:
        """
        return self._data

    def __repr__(self):
        return f"P {[n.id for n in self.nodes]}"

    def set(self, key, val) -> None:
        """

        @param key:
        @param val:
        """
        self._data[key] = val

    def get(self, key, default=None):
        """

        @param key:
        @param default:
        @return:
        """
        try:
            return self._data[key]
        except KeyError:
            return default

    def export(self) -> typing.Dict:
        """

        @return:
        """
        data = self._data.copy()
        data["nodes"] = [n.id for n in self.nodes]
        return data


class PortPath:
    """

    """

    def __init__(self, topology, id, ports: typing.List['Port']):
        self.id = id
        self.topology = topology
        self.ports: typing.List['Port'] = ports
        self._data = {}

    @property
    def local_data(self) -> typing.Dict:
        """

        @return:
        """
        return self._data

    def __repr__(self):
        return f"P {[n.id for n in self.ports]}"

    def set(self, key, val) -> None:
        """

        @param key:
        @param val:
        """
        self._data[key] = val

    def get(self, key, default=None):
        """

        @param key:
        @param default:
        @return:
        """
        try:
            return self._data[key]
        except KeyError:
            return default

    def export(self) -> typing.Dict:
        """

        @return:
        """
        data = self._data.copy()
        data["ports"] = [n.id for n in self.ports]
        return data
